//Use cl::vector instead of STL version
#define __NO_STD_VECTOR
#define __CL_ENABLE_EXCEPTIONS

#include <CL/cl.hpp>
#include <iostream>
#include <fstream>
#include <string>
#include <random>
#include <chrono>
#include <algorithm>
#include <numeric>

using namespace cl;

#include <oclutils.hpp>

int main() {

	try {
#pragma region Initialize GPU

		Context context;
		if (!oclCreateContextBy(context, "intel")) {
			throw Error(CL_INVALID_CONTEXT, "Failed to create a valid context!");
		}

		// Query devices from the context
		vector<Device> devices = context.getInfo<CL_CONTEXT_DEVICES>();

		// Create a command queue and use the first device
		CommandQueue queue(context, devices[0], CL_QUEUE_PROFILING_ENABLE);

		// Read source file
		auto sourceCode = 	oclReadSourcesFromFile("scan_kernels.cl");
		Program::Sources sources(1, std::make_pair(sourceCode.c_str(), sourceCode.length() + 1));

		// Make program of the source code in the context
		Program program(context, sources);

		// Build program for these specific devices
		try {
			program.build(devices);
		}
		catch (Error error) {
			oclPrintError(error);
			std::cerr << "Build Status: " << program.getBuildInfo<CL_PROGRAM_BUILD_STATUS>(devices[0]) << std::endl;
			std::cerr << "Build Options:\t" << program.getBuildInfo<CL_PROGRAM_BUILD_OPTIONS>(devices[0]) << std::endl;
			std::cerr << "Build Log:\t " << program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(devices[0]) << std::endl;
			throw error;
		}

		// Make kernel
		Kernel kernel_naiveScan(program, "naive_scan");

#pragma endregion

		for (unsigned size = 1 << 23; size <= 1 << 28; size = size << 1) {
			std::vector<int> input(size);
			std::vector<int> result(size);
			std::default_random_engine rand(1);
			std::generate(input.begin(), input.end(), [&] () { return rand() % 512 - 256; });

#pragma region Execute kernel

			// Create memory buffers
			Buffer buffer_1(context, CL_MEM_READ_WRITE, size * sizeof(int));
			Buffer buffer_2(context, CL_MEM_READ_WRITE, size * sizeof(int));

			// Copy input to the memory buffer
			queue.enqueueWriteBuffer(buffer_1, CL_TRUE, 0, size * sizeof(int), input.data());
			queue.enqueueWriteBuffer(buffer_2, CL_TRUE, 0, 1 * sizeof(int), input.data());

			double time = 0;

			bool is_even = true;

			// naive scan
			for(unsigned stepSize = 1; stepSize < size; stepSize <<= 1)
			{
				// Run the kernel on specific ND range
				Event operation;

				// "double buffering"
				if (is_even) {
					kernel_naiveScan.setArg(0, buffer_1);
					kernel_naiveScan.setArg(1, buffer_2);
				}
				else {
					kernel_naiveScan.setArg(0, buffer_2);
					kernel_naiveScan.setArg(1, buffer_1);
				}
				is_even = !is_even;

				kernel_naiveScan.setArg(2, stepSize);

				cl::NDRange _global_(size - stepSize);

				queue.enqueueNDRangeKernel(kernel_naiveScan, cl::NullRange, _global_, cl::NullRange, nullptr, &operation);

				// TODO: copy non-processed elements between buffers

				// Measure time
				time += oclGetTiming(operation);
			}

			// Read the result!
			if (!is_even) {
				queue.enqueueReadBuffer(buffer_2, CL_TRUE, 0, size * sizeof(int), &result[0]);
			}
			else {
				queue.enqueueReadBuffer(buffer_1, CL_TRUE, 0, size * sizeof(int), &result[0]);
			}

#pragma endregion

			std::cout << time << std::endl;
			
			for(unsigned i = 1; i<size; ++i)
			{
				//input[i] = input[i-1] + input[i];
				input[i] += input[i - 1];
			}

			for(unsigned i = 0; i<size; ++i)
				if(result[i] != input[i])
				{
					std::cerr << "Error at size: " << size << "; at element: " << i << std::endl;
					throw Error(CL_INVALID_VALUE);
				}
		}

	}
	catch (Error error) {
		oclPrintError(error);
	}

	std::cin.get();

	return 0;
}